## ***********************
##   METHODS - covMatrix
## ***********************

covMatrix.covTensorProduct <- function(object, X, noise.var=NULL) {
	
  d <- ncol(X)
  n <- nrow(X)
  
  param <- covparam2vect(object)
  
  out <- .C("C_covMatrix", 
            as.double(X), as.integer(n), as.integer(d), 
            as.double(param), as.double(object@sd2), as.character(object@name), 
            ans = double(n * n),
            PACKAGE="DiceKriging")
  
  C <- matrix(out$ans, n, n)   # covariance matrix when there is no nugget effect
  
  if (object@nugget.flag) {
    vn <- rep(object@nugget, n)
    C <- C + diag(vn)
  } else if (length(noise.var)>0) {
    vn <- noise.var
    C <- C + diag(noise.var)
  } else {
    vn <- rep(0, n)
  }
    
  return(list(C=C, vn=vn))	
}

covMatrixcovUser <- function(object, X, noise.var=NULL) {
  
  d <- ncol(X)
  n <- nrow(X)
  
  C <- matrix(NA, n, n)
  for(i in 1L:n){
    for(j in i:n){
      C[j,i] <- C[i,j] <- object@kernel(X[i,], X[j,])
    }
  }
    
  if (object@nugget.flag) {
    vn <- rep(object@nugget, n)
    C <- C + diag(vn)
  } else if (length(noise.var)>0) {
    vn <- noise.var
    C <- C + diag(noise.var)
  } else {
    vn <- rep(0, n)
  }
  
  return(list(C=C, vn=vn))  
}

if(!isGeneric("covMatrix")) {
  setGeneric(name    = "covMatrix",
             def     = function(object, X, noise.var=NULL) standardGeneric("covMatrix")
             )
}

setMethod("covMatrix", "covUser", 
          function(object, X, noise.var=NULL) {
            covMatrixcovUser(object=object, X=X, noise.var=noise.var)
          }
          )


setMethod("covMatrix", "covTensorProduct", 
          function(object, X, noise.var=NULL) {
            covMatrix.covTensorProduct(object=object, X=X, noise.var=noise.var)
          }
          )

setMethod("covMatrix", "covIso", 
          function(object, X, noise.var=NULL) {
            covMatrix.covTensorProduct(object=as(object, "covTensorProduct"), X=X, noise.var=noise.var)
          }
          )

setMethod("covMatrix", "covAffineScaling", 
          function(object, X, noise.var=NULL) {
            covMatrix(extract.covIso(object), X=affineScalingFun(X, knots=object@knots, eta=object@eta), noise.var=noise.var)
          }
          )

setMethod("covMatrix", "covScaling", 
          function(object, X, noise.var=NULL) {
            covMatrix(extract.covIso(object), X=scalingFun(X, knots=object@knots, eta=object@eta), noise.var=noise.var)
          }
          )


## *******************************
##   METHODS - covMatrixDerivative
## *******************************

covMatrixDerivative.covTensorProduct <- function(object, X, C, k) {
  ## X : n x d
  n <- nrow(X)
  d <- ncol(X)
  
  ##beware that the index k  starts at 0 in C language 
  
  param <- covparam2vect(object)
	
  out <- .C("C_covMatrixDerivative", 
            as.double(X), as.integer(n), as.integer(d), 
            as.double(param), as.character(object@name),
            as.integer(k), as.double(C),
					ans = double(n * n),
            PACKAGE="DiceKriging")
  
  return(matrix(out$ans, n, n))
}

covMatrixDerivative.dx.covTensorProduct <- function(object, X, C, k) {
  ## X  : n x d
  n <- nrow(X)
  d <- ncol(X)
				
  ## beware that the index k  starts at 0 in C language 
  
  param <- covparam2vect(object)
  
  out <- .C("C_covMatrixDerivative_dx", 
            as.double(X), as.integer(n), as.integer(d), 
            as.double(param), as.character(object@name),
            as.integer(k), as.double(C),
            ans = double(n * n),
            PACKAGE="DiceKriging")
  
  return(matrix(out$ans, n, n))
}


if(!isGeneric("covMatrixDerivative")) {
  setGeneric(name = "covMatrixDerivative",
             def = function(object, X, C, k, ...) standardGeneric("covMatrixDerivative")
             )
}

setMethod("covMatrixDerivative", "covTensorProduct", 
          function(object, X, C, k) {
            covMatrixDerivative.covTensorProduct(object=object, X=X, C=C, k=k)
          }
          )

setMethod("covMatrixDerivative", "covIso", 
          function(object, X, C, k) {
            dC <- matrix(0, nrow(C), ncol(C))
            object <- as(object, "covTensorProduct")
            for (j in 1:object@d) {
              dC <- dC + covMatrixDerivative.covTensorProduct(object=object, X=X, C=C, k=j)
            }
            return(dC)
          }
          )

envir.covScaling <- envir.covAffineScaling <- new.env()

setMethod("covMatrixDerivative", "covAffineScaling", 
          function(object, X, C, k, envir=envir.covAffineScaling) {
            ## NOTE : this function MUST be used in a loop over the index k, from 1 to k.max
            i <- k
            k <- floor((i-1)/2)+1
            l <- (i-1)%%2+1
            if (l==1) {
              object.covTensorProduct <- as(extract.covIso(object), "covTensorProduct")
              fX <- affineScalingFun(X, knots=object@knots, eta=object@eta)
              Dk <- covMatrixDerivative.dx.covTensorProduct(object=object.covTensorProduct, 
                                                            X=fX, C=C, k=k)
              envir$Dk <- Dk
            } else {
              Dk <- envir$Dk
            }
            df.dkl <- affineScalingGrad(X=X, knots=object@knots, k=k)[,l]
            A <- outer(df.dkl, df.dkl, "-")
            return(Dk*A)
          }
          )

setMethod("covMatrixDerivative", "covScaling", 
          function(object, X, C, k, envir=envir.covScaling) {
                                        # NOTE : this function MUST be used in a loop over the index k, from 1 to k.max
            i <- k
            if (i==1) {
              knots.n <- as.numeric(sapply(object@knots, length))
              k.vec <- rep(1:object@d, times=knots.n)
              l.vec <- sequence(knots.n)
              envir$k.vec <- k.vec
              envir$l.vec <- l.vec
            } else {
              k.vec <- envir$k.vec
              l.vec <- envir$l.vec
            }
            k <- k.vec[i]
            l <- l.vec[i]
            
            if (l==1) {
              object.covTensorProduct <- as(extract.covIso(object), "covTensorProduct")
              fX <- scalingFun(X, knots=object@knots, eta=object@eta)
              Dk <- covMatrixDerivative.dx.covTensorProduct(object=object.covTensorProduct, 
                                                            X=fX, C=C, k=k)   
              envir$Dk <- Dk
            } else {
              Dk <- envir$Dk
            }
            df.dkl <- scalingGrad(X=X, knots=object@knots, k)[,l]
            A <- outer(df.dkl, df.dkl, "-")
            return(Dk*A)
          }
          )




## ************************
##   METHODS - covVector.dx
## ************************

covVector.dx.covTensorProduct <- function(object, x, X, c) {

	n <- nrow(X);
	d <- ncol(X);

	param <- covparam2vect(object)

	out <- .C("C_covVector_dx", 
				as.double(x), as.double(X), 
				as.integer(n), as.integer(d),
				as.double(param), as.character(object@name), 
				as.double(c), 
				ans = double(n * d))
   	
   	return(matrix(out$ans, n, d))
	
}

if(!isGeneric("covVector.dx")) {
	setGeneric(name = "covVector.dx",
			   def = function(object, x, X, c) standardGeneric("covVector.dx")
			   )
}

setMethod("covVector.dx", "covTensorProduct", 
function(object, x, X, c) {
	covVector.dx.covTensorProduct(object=object, x=x, X=X, c=c)
}
)

setMethod("covVector.dx", "covIso", 
function(object, x, X, c) {
	covVector.dx.covTensorProduct(object=as(object, "covTensorProduct"), x=x, X=X, c=c)
}
)

# setMethod("covVector.dx", "covAffineScaling",
# function(object, x, X, c) {
#   object.covTensorProduct <- as(extract.covIso(object), "covTensorProduct")
#   fx <- affineScalingFun(matrix(x, nrow=1), knots=object@knots, eta=object@eta)
#   fX <- affineScalingFun(X, knots=object@knots, eta=object@eta)
#   d <- length(x)
#   dkScaling.dx <- rep(Na, d)
#   dk.dx <- covVector.dx.covTensorProduct(object=object.covTensorProduct, x=fx, X=fX, c=c)
#   for (j in 1:d){
#     dkScaling.dx[j] <- dk.dx[j] * affineScalingGrad(X=X, knots=object@knots, k=j)
#   }
#   return(dkScaling.dx)
# }
# )




## ***********************
##   METHODS - covMat1Mat2
## ***********************

covMat1Mat2.covTensorProduct <- function(object, X1, X2, nugget.flag=FALSE) {

	# X1 : matrix n1 x d - containing training points
	# X2 : matrix n2 x d - containing test points
	
	# X1 <- as.matrix(X1)
	# X2 <- checkNames(X1, X2)
	
	n1 <- nrow(X1)
	n2 <- nrow(X2)
	d <- ncol(X1)
	
	param <- covparam2vect(object)
	
	out <- .C("C_covMat1Mat2", 
				as.double(X1), as.integer(n1),
				as.double(X2), as.integer(n2), 
				as.integer(d),
				as.double(param), as.double(object@sd2), as.character(object@name),
				ans = double(n1 * n2), PACKAGE="DiceKriging")
				
	M <- matrix(out$ans, n1, n2)
	
	if ((!nugget.flag) | (!object@nugget.flag)) {
		return(M)
	} else {
		out <- .C("C_covMat1Mat2", 
				as.double(X1), as.integer(n1),
				as.double(X2), as.integer(n2), 
				as.integer(d),
				as.double(param), as.double(object@nugget), "whitenoise",
				ans = double(n1 * n2), PACKAGE="DiceKriging")
		N <- matrix(out$ans, n1, n2)
		return(M+N)
	}
		
}

if(!isGeneric("covMat1Mat2")) {
	setGeneric(name    = "covMat1Mat2",
			   def     = function(object, X1, X2, nugget.flag=FALSE) standardGeneric("covMat1Mat2")
			   )
}

setMethod("covMat1Mat2", "covTensorProduct", 
function(object, X1, X2, nugget.flag=FALSE) {
	covMat1Mat2.covTensorProduct(object=object, X1=X1, X2=X2, nugget.flag=nugget.flag)
}
)

setMethod("covMat1Mat2", "covIso", 
function(object, X1, X2, nugget.flag=FALSE) {
	covMat1Mat2.covTensorProduct(object=as(object, "covTensorProduct"), X1=X1, X2=X2, nugget.flag=nugget.flag)
}
)

setMethod("covMat1Mat2", "covAffineScaling", 
function(object, X1, X2, nugget.flag=FALSE) {
	covMat1Mat2(extract.covIso(object), X1=affineScalingFun(X1, knots=object@knots, eta=object@eta), X2=affineScalingFun(X2, knots=object@knots, eta=object@eta), nugget.flag=nugget.flag)
}
)

setMethod("covMat1Mat2", "covScaling", 
function(object, X1, X2, nugget.flag=FALSE) {
  covMat1Mat2(extract.covIso(object), X1=scalingFun(X1, knots=object@knots, eta=object@eta), X2=scalingFun(X2, knots=object@knots, eta=object@eta), nugget.flag=nugget.flag)
}
)

covMat1Mat2.covUser <- function(object, X1, X2) {

  # X1 : matrix n1 x d - containing training points
	# X2 : matrix n2 x d - containing test points
	
	n1 <- nrow(X1)
	n2 <- nrow(X2)
	d <- ncol(X1)

	M <- matrix(NA, n1, n2)
	for(i in seq(1,n1)){
	for(j in seq(1,n2)){
	M[i,j] <- object@kernel(X1[i,], X2[j,])
	}
	}
				
	return(M)		
}

setMethod("covMat1Mat2", "covUser", 
function(object, X1, X2, nugget.flag=FALSE) {
  M <- covMat1Mat2.covUser(object, X1, X2)
  
  if ((!nugget.flag) | (!object@nugget.flag)) {
  	return(M)
	} else {
    n1 <- nrow(X1)
    n2 <- nrow(X2)
		out <- .C("C_covMat1Mat2", 
				as.double(X1), as.integer(n1),
				as.double(X2), as.integer(n2), 
				as.integer(ncol(X1)),
				as.double(0), as.double(object@nugget), "whitenoise",
				ans = double(n1 * n2), PACKAGE="DiceKriging")
		N <- matrix(out$ans, n1, n2)
		return(M+N) 
  }
}
)

